'PICBasic Plus UDP networking interface for Proton-Net (c) Iosoft Ltd. 2003
'
'This software is supplied 'as-is' for development purposes, without warranty
'of any kind, either expressed or implied, including, but not limited to,
'the implied warranties of merchantability and fitness for purpose.
'
'The purchaser of this software is granted a non-exclusive license to use
'the software for personal experimentation, and in commercial applications,
'providing the Iosoft copyright notice is retained. Executable binary
'images derived from this code may be incorporated into products without
'any license payments, but the source code may not be redistributed. Copies
'of this source code must be purchased from Iosoft Ltd. or a licensed partner.
'For more information, see the Iosoft Ltd. Web site www.iosoft.co.uk  
'
' v0.01 JPB 4/11/02
' v0.02 JPB 6/11/02 ARP Tx working
' v0.03 JPB 6/11/02 Rx working
' v0.04 JPB 6/11/02 Started ARP Rx
' v0.05 JPB 8/11/02 Rearranged variables to fix Rx array index problem
'                   Responds to ARP request
' v0.06 JPB 13/11/02 Responds to pings
'                    (can only handle 22 bytes of data with 64-byte buffer)
' v0.07 JPB 4/12/02  Added client ARP request
' v0.08 JPB 5/12/02  Added UDP time display
' v0.09 JPB 17/12/02 Re-coded LCD display handler
' v0.10 JPB 19/12/02 Added gateway interface and message handler
' v0.11 JPB 19/12/02 Put ADC value in message response
' v0.12 JPB 19/12/02 Added ICMP Destination Unreachable message
' v0.13 JPB 20/12/02 Fixed intermittent ping response (txlen issue)
' v1.01 JPB 8/1/03   Added board-to-board comms
' v1.02 LJ  20/1/03  Moved variables & data into separate files
'                    Replaced FOR-NEXT loops with REPEAT-UNTIL
'                    Restored PRINT command for LCD output
'                    Rearranged subroutine placement, added ADC subroutine
'                    Rearranged variables to optimise BANK0 usage
'                    Replaced AND and OR operands with multiple IF-THEN  
' v1.03 JPB 22/1/03  Updated revision history
'                    Restored primary definitions into this file
'                    Restored default IP addresses
'                    Fixed problem with LCD time display
'                    Re-tested with Picbasic Pro v1.24e

	Device 16f877
    XTAL = 20 	
    SYMBOL CPU_MHZ  = 19660800	' For use with a 19.600 MHz crystal

'------------------------------------------------------------------
'Network definitions
'First 3 bytes of my IP address
	SYMBOL MYIP1    = 10 
	SYMBOL MYIP2    = 1 
	SYMBOL MYIP3    = 1 
'My ID number; last byte of MAC & IP address
	SYMBOL MYID     = 98 

'Port number for time protocol
	SYMBOL TIMEPORT = 37

'Port to be used for incoming messages
	SYMBOL MSGPORT  = 4321

'4 bytes of remote IP address
	SYMBOL REMIP1   = 10 
	SYMBOL REMIP2   = 1 
	SYMBOL REMIP3   = 1 
	SYMBOL REMIP4   = 99 

'Remote port address for client (TIMEPORT or MSGPORT)
	SYMBOL REMPORT  = MSGPORT

'4 bytes of gateway IP address
	SYMBOL GWIP1    = 10 
	SYMBOL GWIP2    = 1 
	SYMBOL GWIP3    = 1 
	SYMBOL GWIP4    = 100   

'3 bytes of IP address mask
	SYMBOL MASKIP1  = 255
	SYMBOL MASKIP2  = 255
	SYMBOL MASKIP3  = 255

	Include "UDP_VARS.INC"				' Add the Variable/Data definitions

'------------------------------------------------------------------
'Default I/O definitions
	PORTA = $ff
	TRISA = $03
	TRISB = $e0
	TRISC = $ff
	TRISD = $ff
	PORTE = $07
	TRISE = $00

	ADCON1 = $8e                		' Only RA0 is analog
	LOW PORTD.5                 		' Pull the LCD's RW line LOW
	CLS
	PRINT "PB_UDP  v1.03", AT 2,1, DEC MYIP1, ".", DEC MYIP2, ".", DEC MYIP3, ".", DEC MYID
	INPUT PORTD

	HRSOUT "PB_UDP v1.03", $d, $a,_
	"My IP ", DEC MYIP1, ".", DEC MYIP2, ".", DEC MYIP3, ".", DEC MYID,_
	", host ", DEC REMIP1, ".", DEC REMIP2, ".", DEC REMIP3, ".", DEC REMIP4,_
	", gate ", DEC GWIP1, ".", DEC GWIP2, ".", DEC GWIP3, ".", DEC GWIP4, $d, $a
	
    rem_ip[0] = REMIP1: rem_ip[1] = REMIP2: rem_ip[2] = REMIP3: rem_ip[3] = REMIP4
	gw_ip[0] = GWIP1: gw_ip[1] = GWIP2: gw_ip[2] = GWIP3: gw_ip[3] = GWIP4
    
    T1CON = $31
	tickcount = 0
    ledtickcount = 0
    arped = 0
    
    GOTO start							'Jump around subroutines
'------------------------------------------------------------------
' Display the contents of a portion of rxbuff array
DISPLAY_LINE:
	idx = UDP_DATA_START            	'Display on 2nd LCD line
    OUTPUT PORTD
    CURSOR 2,1                	
    i = 0
    REPEAT								'16 chars on LCD
    	b = rxbuff[idx]
    	IF idx < rxlen AND b > 0 THEN   'Maybe fewer in message
    		PRINT b
    		INC idx
    	ELSE
    		PRINT " "
    	ENDIF
    INC i
    UNTIL i > 15
    INPUT PORTD
RETURN
'------------------------------------------------------------------
'Do an NIC write cycle
out_nic:
    OUTPUT PORTD
    NIC_IOW = 0
    DELAYUS 3
    NIC_IOW = 1
    INPUT PORTD
    INC txlen
RETURN
'------------------------------------------------------------------
'Do an NIC read cycle
in_nic:
    INPUT PORTD
    NIC_IOR = 0
    DELAYUS 3
    nicd = NICDATA
    NIC_IOR = 1
RETURN
'------------------------------------------------------------------
'Check the data length for a given task, and set it in the NIC
get_datalen:
    IF task > NIC_INIT THEN RETURN
    IF task < TASKBASE THEN RETURN
    RESTORE
    REPEAT                                      'Search for task marker
        READ a: READ b
        IF a = T_END THEN RETURN
    UNTIL a = task
    datalen = 0
    REPEAT                                      'Tally data lengths
        READ a: READ b
        IF a = DPORT THEN datalen = datalen + 1
        IF a = T_VOUT2 THEN datalen = datalen + 2
        IF a = T_VOUT4 THEN datalen = datalen + 4
        IF a = T_VOUT6 THEN datalen = datalen + 6
    UNTIL a >= T_STOP
    NICADDR = RBCR0: NICDATA = datalen: GOSUB out_nic
RETURN
'------------------------------------------------------------------
'Output the variable specified in 'b' to the NIC transmit buffer
vout:
    NICADDR = DPORT
    IF b < V_BASE THEN
        IF a = T_VOUT1 THEN                         '1 byte from rx buffer
            NICDATA = rxbuff[b]: GOSUB out_nic
        ENDIF
        IF a = T_VOUT2 THEN                         '2 bytes from rx buffer
            NICDATA = rxbuff[b]: GOSUB out_nic
            NICDATA = rxbuff[b + 1]: GOSUB out_nic
        ENDIF
        IF a = T_VOUT4 THEN                         '4 bytes from rx buffer
            i = 0: REPEAT: NICDATA = rxbuff[b + i]: GOSUB out_nic: INC i: UNTIL i > 3
        ENDIF
        IF a = T_VOUT6 THEN                         '6 bytes from rx buffer
            i = 0: REPEAT: NICDATA = rxbuff[b + i]: GOSUB out_nic: INC i: UNTIL i > 5
        ENDIF
    ENDIF
    IF b = V_LOC_MAC THEN                           'Local MAC address
        NICDATA = $02: GOSUB out_nic
        NICDATA = $00: GOSUB out_nic
        NICDATA = $00: GOSUB out_nic
        NICDATA = $00: GOSUB out_nic
        NICDATA = $00: GOSUB out_nic
        NICDATA = MYID: GOSUB out_nic
    ENDIF
    IF b = V_LOC_IP THEN                            'Local IP address
        NICDATA = MYIP1: GOSUB out_nic
        NICDATA = MYIP2: GOSUB out_nic
        NICDATA = MYIP3: GOSUB out_nic
        NICDATA = MYID: GOSUB out_nic
    ENDIF
    IF b = V_REM_MAC THEN                           'Remote MAC address
        I = 0: REPEAT: NICDATA = rem_mac[i]: GOSUB out_nic: INC I: UNTIL I > MACLEN - 1
    ENDIF
    IF b = V_RESP_MAC THEN                          'Response MAC address
        I = 0: REPEAT: NICDATA = rxbuff[MACLEN + i]: GOSUB out_nic: INC I: UNTIL I > MACLEN - 1
    ENDIF
    a = ((REMIP1 ^ MYIP1) & MASKIP1) | ((REMIP2 ^ MYIP2) & MASKIP2) | ((REMIP3 ^ MYIP3) & MASKIP3)
    IF b = V_REMGATE_IP AND a <> 0 THEN             'Gateway IP address
        NICDATA = GWIP1: GOSUB out_nic              '(only if off subnet)
        NICDATA = GWIP2: GOSUB out_nic
        NICDATA = GWIP3: GOSUB out_nic
        NICDATA = GWIP4: GOSUB out_nic
        RETURN
    ENDIF
    IF b = V_REM_IP OR b = V_REMGATE_IP THEN        'Remote IP address
        I = 0: REPEAT: NICDATA = rem_ip[i]: GOSUB out_nic: INC I: UNTIL I > IPADDRLEN - 1
    ENDIF
    IF b = V_IPLEN THEN                             'IP length word
        NICDATA = 0: GOSUB out_nic
        NICDATA = iplen: GOSUB out_nic
    ENDIF
    IF b = V_UDPLEN THEN                            'UDP length word
        NICDATA = 0: GOSUB out_nic
        NICDATA = udplen: GOSUB out_nic
    ENDIF
    IF b = V_VOLTAGE THEN                           '4-char ADC voltage
        NICDATA = "0" + DIG centivolts, 2: GOSUB out_nic
        NICDATA = ".": GOSUB out_nic
        NICDATA = "0" + DIG centivolts, 1: GOSUB out_nic
        NICDATA = "0" + DIG centivolts, 0: GOSUB out_nic
    ENDIF
RETURN
'------------------------------------------------------------------
'Execute a given task on the NIC
do_nic_task:
    IF task > NIC_INIT THEN RETURN
    IF task < TASKBASE THEN RETURN
    RESTORE
    REPEAT                                      			'Search for task marker
        READ a: READ b
        IF a = T_END THEN RETURN
    UNTIL a = task
    REPEAT                                      			'Read byte pair
        READ a: READ b                              		'Can be register & value
        IF a <= RESPORT THEN NICADDR = a: NICDATA = b: GOSUB out_nic
        IF a = T_WAIT THEN DELAYMS b                		'..or delay value
        IF a >= T_VOUT1 THEN IF a <= T_VOUT6 THEN GOSUB vout	'..or variable & value
    UNTIL a >= T_STOP
RETURN
'------------------------------------------------------------------
'Match my IP address, given index into the Rx buffer
match_myip:
    matched = 0
    IF rxlen >= idx + 4 THEN
    	IF rxbuff[idx] = MYIP1 THEN
    		IF rxbuff[idx + 1] = MYIP2 THEN
        		IF rxbuff[idx + 2] = MYIP3 THEN IF rxbuff[idx + 3] = MYID THEN matched = 1
    		ENDIF
    	ENDIF
    ENDIF
RETURN
'------------------------------------------------------------------
'Match remote IP address, given index into the Rx buffer
match_remip:
    matched = 0
    IF rxlen >= idx + 4 THEN
    	IF rxbuff[idx] = REMIP1 THEN
    		IF rxbuff[idx + 1] = REMIP2 THEN
        		IF rxbuff[idx + 2] = REMIP3 THEN IF rxbuff[idx + 3] = REMIP4 THEN matched = 1
    		ENDIF
        ENDIF
    ENDIF
RETURN
'------------------------------------------------------------------
'Match gateway IP address, given index into the Rx buffer
match_gwip:
    IF rxlen >= idx + 4 THEN
    	IF rxbuff[idx] = GWIP1 THEN
    		IF rxbuff[idx + 1] = GWIP2 THEN
        		IF rxbuff[idx + 2] = GWIP3 THEN IF rxbuff[idx + 3] = GWIP4 THEN matched = 1
            ENDIF
        ENDIF
    ENDIF
RETURN
'------------------------------------------------------------------
'Get remote MAC address, given index into the Rx buffer
get_remmac:
    i = 0
    REPEAT
        rem_mac[i] = rxbuff[idx + i]
        INC i
    UNTIL i > MACLEN - 1
RETURN
'------------------------------------------------------------------
'Add a byte to the checksum
'Use 16-bit addition, with carry from M.S.bit to L.S.bit
csum_byte:
    IF csumf = 1 THEN
        csuml = csuml  +  nicd
        IF csuml < nicd THEN
            INC csumh
            IF csumh = 0 THEN Inc csuml
        ENDIF
        csumf = 0
    ELSE
        csumh = csumh + nicd
        IF csumh < nicd THEN
            INC csuml
            IF csuml = 0 THEN Inc csumh
        ENDIF
        csumf = 1
    ENDIF
RETURN
'------------------------------------------------------------------
'Calculate the Tx IP checksum
calc_tx_ip_csum:
    csuml = 0: csumh = 0: csumf = 0
    NICADDR = RSAR0: NICDATA = IPHDR_START: GOSUB out_nic
    NICADDR = RSAR1: NICDATA = TXSTART: GOSUB out_nic
    NICADDR = RBCR0: NICDATA = 20: GOSUB out_nic
    NICADDR = RBCR1: NICDATA = 0: GOSUB out_nic
    NICADDR = CMDR: NICDATA = $0a: GOSUB out_nic
    NICADDR = DPORT
    i = 0: REPEAT: GOSUB in_nic: GOSUB csum_byte: INC i: UNTIL i > 19 
    NICADDR = RSAR0: NICDATA = 24: GOSUB out_nic
    NICADDR = RSAR1: NICDATA = TXSTART: GOSUB out_nic
    NICADDR = RBCR0: NICDATA = 2: GOSUB out_nic
    NICADDR = RBCR1: NICDATA = 0: GOSUB out_nic
    NICADDR = CMDR: NICDATA = $12: GOSUB out_nic
    NICADDR = DPORT: NICDATA = ~csumh: GOSUB out_nic
    NICADDR = DPORT: NICDATA = ~csuml: GOSUB out_nic
RETURN
'------------------------------------------------------------------
'Transmit the frame in the NIC
transmit:
    IF txlen < MINFRAME THEN txlen = MINFRAME
    NICADDR = TBCR0: NICDATA = txlen: GOSUB out_nic
    NICADDR = TBCR1: NICDATA = 0: GOSUB out_nic
    NICADDR = CMDR: NICDATA = $24: GOSUB out_nic
RETURN
'------------------------------------------------------------------
'Fetch any incoming network data
receive:
    rxlen = 0
    txlen = 0
    NICADDR = ISR: GOSUB in_nic
    IF (nicd & $10) = $10 THEN
        HRSOUT "Rx overrun", $d, $a
        task = NIC_INIT: GOSUB do_nic_task
        RETURN
    ENDIF
    NICADDR = CMDR: NICDATA = $60 : GOSUB out_nic   'Get current NIC page
    NICADDR = CURRP: GOSUB in_nic
    NICADDR = CMDR: NICDATA = $20 : GOSUB out_nic
    IF nicd <> next_page THEN                       'If changed..
        curr_page = next_page
        NICADDR = ISR: NICDATA = $40 : GOSUB out_nic    'Clear ISR
        NICADDR = RSAR0: NICDATA = 0 : GOSUB out_nic    'Get 4-byte page hdr
        NICADDR = RSAR1: NICDATA = curr_page : GOSUB out_nic
        NICADDR = RBCR0: NICDATA = 4 : GOSUB out_nic
        NICADDR = RBCR1: NICDATA = 0 : GOSUB out_nic
        NICADDR = CMDR: NICDATA = $0a : GOSUB out_nic
        NICADDR = DPORT
        i = 0: REPEAT: GOSUB in_nic: rxstat[i] = nicd: INC i: UNTIL i > 3
        IF (rxstat[0] & 1) = 1 AND rxblen >= MINFRAMEC AND rxblen <= MAXFRAMEC THEN
            NICADDR = RBCR0: NICDATA = RXBUFFLEN : GOSUB out_nic
            NICADDR = RBCR1: NICDATA = 0 : GOSUB out_nic    'Get frame data
            NICADDR = CMDR: NICDATA = $0a : GOSUB out_nic
            NICADDR = DPORT
            rxblen = rxblen - 4
            IF rxblen > RXBUFFLEN THEN rxblen = RXBUFFLEN
            rxlen = rxblen
            i = 0
            REPEAT
                GOSUB in_nic: rxbuff[i] = nicd;
            	INC i
            UNTIL i > rxlen
        ELSE
            HRSOUT "NIC frame error", $d, $a
        ENDIF
        IF rxstat#1 >= RXSTART AND rxstat#1 < RXSTOP THEN
            next_page = rxstat#1                   'Update ptr to next page
        ELSE
            INC next_page
            IF next_page >= RXSTOP THEN next_page = next_page + RXSTART - RXSTOP
        ENDIF
        DEC curr_page
        IF curr_page < RXSTART THEN curr_page = curr_page + RXSTOP - RXSTART
        NICADDR = BNRY: NICDATA = curr_page : GOSUB out_nic
    ENDIF
RETURN
'------------------------------------------------------------------
'Check the received network frame
check_rx:
    IF pcol = PCOL_ARP THEN
        idx = ETHLEN + 24 : GOSUB match_myip
        IF matched = 1 THEN
            IF arp_op = $0001 THEN                  'ARP request
                HRSOUT "ARP req"
                txlen = 0                           'Send ARP response
                task = TX_ARPRSP: GOSUB get_datalen: GOSUB do_nic_task
                GOSUB transmit
            ENDIF
            IF arp_op = $0002 THEN                  'ARP response
                idx = ETHLEN + 14: GOSUB match_remip: GOSUB match_gwip
                IF matched = 1 THEN
                    HRSOUT "ARP resp"               'Get MAC address
                    idx = ETHLEN + 8: GOSUB get_remmac
                    arped = 1
                ENDIF
            ENDIF
        ENDIF
    ENDIF
    IF pcol = PCOL_IP THEN                          'IP protocol
        idx = 30 : GOSUB match_myip                 'Match my address
        IF matched = 1 THEN
            HRSOUT "IP "
            IF rxlen > rxiplen + MACHDR_LEN THEN rxlen = rxiplen + MACHDR_LEN
            IF ip_pcol = 1 THEN                     'ICMP protocol
                HRSOUT "ICMP "
                IF icmp_type = 8 THEN IF rxiplen < 256 THEN
                    iplen = rxiplen                 'ICMP echo request
                    task = TX_ICMPRSP: GOSUB get_datalen: GOSUB do_nic_task
                    icmp_type = 0                   'Send echo response
                    icmp_csum = icmp_csum + 8       'Adjust checksum
                    IF icmp_csum < 8 THEN icmp_csum = icmp_csum + 1
                    NICADDR = RBCR0: NICDATA = iplen - 20: GOSUB out_nic
                    NICADDR = RBCR1: NICDATA = 0 : GOSUB out_nic
                    NICADDR = CMDR: NICDATA = $12 : GOSUB out_nic
                    NICADDR = DPORT                 'Copy header & data
                    i = 0
                    REPEAT
                        NICDATA = rxbuff[i + 34]: GOSUB out_nic
                    	INC i
                    UNTIL i > iplen - 21
                    GOSUB calc_tx_ip_csum           'Do IP checksum
                    GOSUB transmit                  'Send the frame
                ENDIF
                IF icmp_type = 3 THEN               'ICMP dest unreachable
                    HRSOUT "Destination unreachable"
                ENDIF
            ENDIF
            IF ip_pcol = 17 THEN                    'UDP datagram
                HRSOUT "UDP "
                IF rx_srce_port = TIMEPORT
                    HRSOUT "time "                  	'Response to time request
                    rx_udp_dw = rx_udp_dw - $7b98a000 	'Subtract days to make +ve
                    s = rx_udp_dw // 60             	'Get hrs, min, sec
                    rx_udp_dw = rx_udp_dw / 60
                    m = rx_udp_dw // 60
                    rx_udp_dw = rx_udp_dw / 60
                    h = rx_udp_dw // 24
                    'Display on top LCD line 	
                    OUTPUT PORTD 
                    Print $fe , $88 , DEC2 h , ":" , DEC2 m , ":" , DEC2 s
                    INPUT PORTD 
                ENDIF
                IF rx_dest_port = MSGPORT
                    HRSOUT "message req "           	'Incoming text message
                    GOSUB DISPLAY_LINE					'Display on 2nd LCD line
                    'Respond with ADC value..
                    centivolts = adcval ** 32000        'Get ADC volts / 100
                    task = TX_MSGRSP: GOSUB get_datalen 'Get response length
                    iplen = datalen - MACHDR_LEN        'Calc IP & UDP len
                    udplen = iplen - IPHDR_LEN
                    GOSUB do_nic_task                   'Send response
                    GOSUB calc_tx_ip_csum: GOSUB transmit
                ENDIF
                IF rx_srce_port = MSGPORT
                    HRSOUT "message resp "          	'Incoming message response           	
                    GOSUB DISPLAY_LINE					'Display on 2nd LCD line
                ENDIF
            ENDIF
        ENDIF
    ENDIF
RETURN
'------------------------------------------------------------------
'Send a client request
update_clock:
    IF arped = 0 THEN
        HRSOUT "Tx ARP request", $d, $a
        txlen = 0
        task = TX_ARPREQ: GOSUB get_datalen: GOSUB do_nic_task
        GOSUB transmit
    ELSE
        HRSOUT "Tx client request", $d, $a
        iplen = 28: txlen = 0
        task = TX_CLIENTREQ: GOSUB get_datalen: GOSUB do_nic_task
        GOSUB calc_tx_ip_csum
        GOSUB transmit
    ENDIF
RETURN
'------------------------------------------------------------------
'Start of main code
start:
    NIC_RST = 1 : NIC_IOR = 1 : NIC_IOW = 1 : LCD_E = 0
    h = 0: m = 0: s = 0
    ULED1 = 1
    DELAYMS 1
    NIC_RST = 0
    DELAYMS 10
    task = NIC_INIT: GOSUB do_nic_task          'Initialise NIC
    next_page = RXSTART + 1
    WHILE 1 = 1                                 'Loop forever..
        tc = TMR1H - lastc
        IF tc >= TIMER1_DIV THEN                'If tick timeout..
            adcval = ADIN 0                     'Get ADC value
            lastc = lastc + TIMER1_DIV
            INC tickcount           			'Update tick count
            tc = tickcount - ledtickcount
            IF tc >= LEDTICK_DIV THEN           'If 1-second timeout..
                GOSUB update_clock              'Get new clock value
                ledtickcount = ledtickcount + LEDTICK_DIV
                TOGGLE SYSLED					'Toggle SYSTEM LED
            ENDIF
        ENDIF
        GOSUB receive                           'Check for Rx frame
        IF rxlen > 0 THEN                       'If frame has arrived..
            HRSOUT "Rx len ", DEC rxlen, " pcol ", HEX pcol, " "
            GOSUB check_rx                      'Decode frame
            HRSOUT $d, $a
        ENDIF
    WEND


